@@ -10,6 +10,10 @@ import datetime
 import irods.password_obfuscation as obf
 from irods import MAX_NAME_LEN
 from ast import literal_eval as safe_eval
+import re
+
+
+PAM_PW_ESC_PATTERN = re.compile(r'([@=&;])')
 
 
 from irods.message import (
@@ -21,7 +25,9 @@ from irods.message import (PamAuthRequest, PamAuthRequestOut)
 
 
 ALLOW_PAM_LONG_TOKENS = True      # True to fix [#279]
-
+# Message to be logged when the connection
+# destructor is called. Used in a unit test
+DESTRUCTOR_MSG = "connection __del__() called"
 
 from irods import (
     MAX_PASSWORD_LENGTH, RESPONSE_LEN,
@@ -80,8 +86,8 @@ class Connection(object):
         return self._client_signature
 
     def __del__(self):
-        if self.socket and getattr(self,"_disconnected",False):
-            self.disconnect()
+        self.disconnect()
+        logger.debug(DESTRUCTOR_MSG)
 
     def send(self, message):
         string = message.pack()
@@ -107,7 +113,12 @@ class Connection(object):
                 msg = iRODSMessage.recv(self.socket)
             else:
                 msg = iRODSMessage.recv_into(self.socket, into_buffer)
-        except socket.error:
+        except (socket.error, socket.timeout) as e:
+            # If _recv_message_in_len() fails in recv() or recv_into(),
+            # it will throw a socket.error exception. The exception is
+            # caught here, a critical message is logged, and is wrapped
+            # in a NetworkException with a more user friendly message
+            logger.critical(e)
             logger.error("Could not receive server response")
             self.release(True)
             raise NetworkException("Could not receive server response")
@@ -273,17 +284,28 @@ class Connection(object):
         return version_msg.get_main_message(VersionResponse)
 
     def disconnect(self):
-        disconnect_msg = iRODSMessage(msg_type='RODS_DISCONNECT')
-        self.send(disconnect_msg)
+        # Moved the conditions to call disconnect() inside the function.
+        # Added a new criteria for calling disconnect(); Only call
+        # disconnect() if fileno is not -1 (fileno -1 indicates the socket
+        # is already closed). This makes it safe to call disconnect multiple
+        # times on the same connection. The first call cleans up the resources
+        # and next calls are no-ops
         try:
-            # SSL shutdown handshake
-            self.socket = self.socket.unwrap()
-        except AttributeError:
-            pass
-        self.socket.shutdown(socket.SHUT_RDWR)
-        self.socket.close()
-        self.socket = None
-        self._disconnected = True
+            if self.socket and getattr(self, "_disconnected", False) == False and self.socket.fileno() != -1:
+                disconnect_msg = iRODSMessage(msg_type='RODS_DISCONNECT')
+                self.send(disconnect_msg)
+                try:
+                    # SSL shutdown handshake
+                    self.socket = self.socket.unwrap()
+                except AttributeError:
+                    pass
+                self.socket.shutdown(socket.SHUT_RDWR)
+                self.socket.close()
+        finally:
+            self._disconnected = True  # Issue 368 - because of undefined destruction order during interpreter shutdown,
+            self.socket = None         # as well as the fact that unhandled exceptions are ignored in __del__, we'd at least
+                                       # like to ensure as much cleanup as possible, thus preventing the above socket shutdown
+                                       # procedure from running too many times and creating confusing messages
 
     def recvall(self, n):
         # Helper function to recv n bytes or return None if EOF is hit
@@ -388,7 +410,7 @@ class Connection(object):
             username=self.account.proxy_user + '#' + self.account.proxy_zone
         )
         gsi_request = iRODSMessage(
-            msg_type='RODS_API_REQ', int_info=704, msg=gsi_msg)
+            msg_type='RODS_API_REQ', int_info=api_number['AUTH_RESPONSE_AN'], msg=gsi_msg)
         self.send(gsi_request)
         self.recv()
         # auth_response = self.recv()
@@ -411,8 +433,10 @@ class Connection(object):
 
         time_to_live_in_seconds = 60
 
+        pam_password = PAM_PW_ESC_PATTERN.sub(lambda m: '\\'+m.group(1), self.account.password)
+
         ctx_user = '%s=%s' % (AUTH_USER_KEY, self.account.client_user)
-        ctx_pwd = '%s=%s' % (AUTH_PWD_KEY, self.account.password)
+        ctx_pwd = '%s=%s' % (AUTH_PWD_KEY, pam_password)
         ctx_ttl = '%s=%s' % (AUTH_TTL_KEY, str(time_to_live_in_seconds))
 
         ctx = ";".join([ctx_user, ctx_pwd, ctx_ttl])
@@ -427,7 +451,7 @@ class Connection(object):
 
             message_body = PamAuthRequest(
                 pamUser=self.account.client_user,
-                pamPassword=self.account.password,
+                pamPassword=pam_password,
                 timeToLive=time_to_live_in_seconds)
         else:
 
@@ -493,7 +517,7 @@ class Connection(object):
 
         # Default case, PAM login will send a new password
         if password is None:
-            password = self.account.password
+            password = self.account.password or ''
 
         # authenticate
         auth_req = iRODSMessage(msg_type='RODS_API_REQ', int_info=703)
@@ -532,10 +556,11 @@ class Connection(object):
             encoded_pwd_array = bytearray(encoded_pwd)
             encoded_pwd = bytes(encoded_pwd_array.replace(b'\x00', b'\x01'))
 
+
         pwd_msg = AuthResponse(
             response=encoded_pwd, username=self.account.proxy_user)
         pwd_request = iRODSMessage(
-            msg_type='RODS_API_REQ', int_info=704, msg=pwd_msg)
+            msg_type='RODS_API_REQ', int_info=api_number['AUTH_RESPONSE_AN'], msg=pwd_msg)
         self.send(pwd_request)
         self.recv()
 
