@@ -554,6 +554,18 @@ class SYS_INVALID_INPUT_PARAM(SystemException):
     code = -130000
 
 
+class SYS_BAD_INPUT(iRODSException):
+    code = -158000
+
+
+class SYS_REPLICA_DOES_NOT_EXIST(iRODSException):
+    code = -164000
+
+
+class SYS_NOT_ALLOWED(iRODSException):
+    code = -169000
+
+
 class UserInputException(iRODSException):
     pass
 
@@ -746,10 +758,18 @@ class NO_LOCAL_FILE_RSYNC_IN_MSI(UserInputException):
     code = -356000
 
 
+class OBJ_PATH_DOES_NOT_EXIST(iRODSException):
+    code = -358000
+
+
 class LOCKED_DATA_OBJECT_ACCESS(SystemException):
     code = -406000
 
 
+class USER_INCOMPATIBLE_PARAMS(iRODSException):
+    code = -402000
+
+
 class CHECK_VERIFICATION_RESULTS(SystemException):
     code = -407000
 
